from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb_sql.parser import ast

import pandas as pd


class CoinBaseAggregatedTradesTable(APITable):

    DEFAULT_INTERVAL = 60
    DEFAULT_SYMBOL = 'BTC-USD'

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Selects data from the CoinBase API and returns it as a pandas DataFrame.

        Returns dataframe representing the CoinBase API results.

        Args:
            query (ast.Select): Given SQL SELECT query
        """
        conditions = extract_comparison_conditions(query.where)

        params = {
            'interval': CoinBaseAggregatedTradesTable.DEFAULT_INTERVAL,
            'symbol': CoinBaseAggregatedTradesTable.DEFAULT_SYMBOL,
        }
        for op, arg1, arg2 in conditions:
            if arg1 == 'interval':
                if op != '=':
                    raise NotImplementedError
                params['interval'] = arg2

            elif arg1 == 'symbol':
                if op != '=':
                    raise NotImplementedError
                params['symbol'] = arg2

        coinbase_candle_data = self.handler.call_coinbase_api(
            method_name='get_candle',
            params=params
        )

        return coinbase_candle_data

    def get_columns(self):
        """Gets all columns to be returned in pandas DataFrame responses"""
        return [
            'symbol',
            'low',
            'high',
            'open',
            'close',
            'volume',
            'timestamp',
            'current_time'
        ]



"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet


class ColumnCountPacket(Packet):
    def setup(self):
        count = self._kwargs.get('count', 0)
        self.column_count = Datum('int<lenenc>', count)

    @property
    def body(self):

        order = [
            'column_count'
        ]

        string = b''
        for key in order:
            string += getattr(self, key).toStringPacket()

        self.setBody(string)
        return self._body

    @staticmethod
    def test():
        import pprint
        pprint.pprint(
            str(ColumnCountPacket(count=1).get_packet_string())
        )


if __name__ == "__main__":
    ColumnCountPacket.test()



import base64
import secrets
import time
import urllib

import requests
from flask import redirect, request, session, url_for
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.auth import ns_conf
from mindsdb.utilities.config import Config
from mindsdb.utilities import log

logger = log.getLogger(__name__)


def get_access_token() -> str:
    """return current access token

    Returns:
        str: token
    """
    return (
        Config().get("auth", {}).get("oauth", {}).get("tokens", {}).get("access_token")
    )


def request_user_info(access_token: str = None) -> dict:
    """request user info from cloud

    Args:
        access_token (str, optional): token that used to get user data

    Returns:
        dict: user data
    """
    if access_token is None:
        access_token = get_access_token()
    if access_token is None:
        raise KeyError()

    auth_server = Config()["auth"]["oauth"]["server_host"]

    response = requests.get(
        f"https://{auth_server}/auth/userinfo",
        headers={"Authorization": f"Bearer {access_token}"},
        timeout=5,
    )
    if response.status_code != 200:
        raise Exception(f"Wrong response: {response.status_code}, {response.text}")

    return response.json()


@ns_conf.route("/callback", methods=["GET"])
@ns_conf.route("/callback/cloud_home", methods=["GET"])
@ns_conf.hide
class Auth(Resource):
    @ns_conf.doc(params={"code": "authentification code"})
    def get(self):
        """callback from auth server if authentification is successful"""
        config = Config()
        code = request.args.get("code")

        aws_meta_data = config["aws_meta_data"]
        public_hostname = aws_meta_data["public-hostname"]
        instance_id = aws_meta_data["instance-id"]

        oauth_meta = config["auth"]["oauth"]
        client_id = oauth_meta["client_id"]
        client_secret = oauth_meta["client_secret"]
        auth_server = oauth_meta["server_host"]
        client_basic = base64.b64encode(
            f"{client_id}:{client_secret}".encode()
        ).decode()

        redirect_uri = f"https://{public_hostname}{request.path}"
        response = requests.post(
            f"https://{auth_server}/auth/token",
            data={
                "code": code,
                "grant_type": "authorization_code",
                "redirect_uri": redirect_uri,
            },
            headers={"Authorization": f"Basic {client_basic}"},
        )
        tokens = response.json()
        if "expires_in" in tokens:
            tokens["expires_at"] = round(time.time() + tokens["expires_in"] - 1)
            del tokens["expires_in"]

        user_data = request_user_info(tokens["access_token"])

        previous_username = config["auth"]["oauth"].get("username")
        new_username = user_data["name"]
        if previous_username is not None and new_username != previous_username:
            return redirect("/forbidden")

        config.update(
            {
                "auth": {
                    "provider": "cloud",
                    "oauth": {"username": new_username, "tokens": tokens},
                }
            }
        )

        try:
            resp = requests.put(
                f"https://{auth_server}/cloud/instance",
                json={
                    "instance_id": instance_id,
                    "public_hostname": public_hostname,
                    "ami_id": aws_meta_data.get("ami-id"),
                },
                headers={"Authorization": f'Bearer {tokens["access_token"]}'},
                timeout=5,
            )
            if resp.status_code != 200:
                logger.warning(f"Wrong response from cloud server: {resp.status_code}")
        except Exception as e:
            logger.warning(f"Cant't send request to cloud server: {e}")

        session["username"] = user_data["name"]
        session["auth_provider"] = "cloud"
        session.permanent = True

        if request.path.endswith("/auth/callback/cloud_home"):
            return redirect(f"https://{auth_server}")
        else:
            return redirect(url_for("root_index"))


@ns_conf.route("/cloud_login", methods=["GET"])
@ns_conf.hide
class CloudLoginRoute(Resource):
    @ns_conf.doc(
        responses={302: "Redirect to auth server"},
        params={"location": "final redirection should lead to that location"},
    )
    def get(self):
        """redirect to cloud login form"""
        location = request.args.get("location")
        config = Config()

        aws_meta_data = config["aws_meta_data"]
        public_hostname = aws_meta_data["public-hostname"]
        auth_server = config["auth"]["oauth"]["server_host"]

        if location == "cloud_home":
            redirect_uri = f"https://{public_hostname}/api/auth/callback/cloud_home"
        else:
            redirect_uri = f"https://{public_hostname}/api/auth/callback"

        args = urllib.parse.urlencode(
            {
                "client_id": config["auth"]["oauth"]["client_id"],
                "scope": "openid profile aws_marketplace",
                "response_type": "code",
                "nonce": secrets.token_urlsafe(),
                "redirect_uri": redirect_uri,
            }
        )
        return redirect(f"https://{auth_server}/auth/authorize?{args}")



from typing import Optional
from collections import OrderedDict

import json
import time
import requests
import pandas as pd

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.base import DatabaseHandler
from sqlalchemy_dremio.base import DremioDialect

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

logger = log.getLogger(__name__)


class DremioHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Dremio statements.
    """

    name = 'dremio'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'dremio'

        self.connection_data = connection_data
        self.kwargs = kwargs

        self.base_url = f"http://{self.connection_data['host']}:{self.connection_data['port']}"

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> dict:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        headers = {
            'Content-Type': 'application/json',
        }

        data = '{' + f'"userName": "{self.connection_data["username"]}","password": "{self.connection_data["password"]}"' + '}'

        response = requests.post(self.base_url + '/apiv2/login', headers=headers, data=data)

        return {
            'Authorization': '_dremio' + response.json()['token'],
            'Content-Type': 'application/json',
        }

    def disconnect(self):
        """
        Close any existing connections.
        """

        self.is_connected = False
        return

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Dremio, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        query = query.replace('"', '\\"').replace('\n', ' ')

        need_to_close = self.is_connected is False

        auth_headers = self.connect()
        data = '{' + f'"sql": "{query}"' + '}'

        try:
            sql_result = requests.post(self.base_url + '/api/v3/sql', headers=auth_headers, data=data)

            job_id = sql_result.json()['id']

            if sql_result.status_code == 200:
                logger.info('Job creation successful. Job id is: ' + job_id)
            else:
                logger.info('Job creation failed.')

            logger.info('Waiting for the job to complete...')

            job_status = requests.request("GET", self.base_url + "/api/v3/job/" + job_id, headers=auth_headers).json()[
                'jobState']

            while job_status != 'COMPLETED':
                time.sleep(2)
                job_status = requests.request("GET", self.base_url + "/api/v3/job/" + job_id, headers=auth_headers).json()[
                    'jobState']

            job_result = json.loads(requests.request("GET", self.base_url + "/api/v3/job/" + job_id + "/results", headers=auth_headers).text)

            if 'errorMessage' not in job_result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        job_result['rows']
                    )
                )
            else:
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(job_result['errorMessage'])
                )

        except Exception as e:
            logger.error(f'Error running query: {query} on Dremio!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        renderer = SqlalchemyRender(DremioDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        query = 'SELECT * FROM INFORMATION_SCHEMA.\\"TABLES\\"'
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        query = f"DESCRIBE {table_name}"
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={'COLUMN_NAME': 'column_name', 'DATA_TYPE': 'data_type'})
        return result


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Dremio server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The port that Dremio is running on.'
    },
    username={
        'type': ARG_TYPE.STR,
        'description': 'The username used to authenticate with the Dremio server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Dremio server.'
    }
)

connection_args_example = OrderedDict(
    host='localhost',
    database=9047,
    username='admin',
    password='password'
)



from collections import OrderedDict
from typing import List, Optional

import lancedb
import pandas as pd
import pyarrow as pa
from lance.vector import vec_to_table
import duckdb
import json

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.response import HandlerResponse as Response
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    FilterOperator,
    TableField,
    VectorStoreHandler,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class LanceDBHandler(VectorStoreHandler):
    """This handler handles connection and execution of the LanceDB statements."""

    name = "lancedb"

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self._connection_data = kwargs.get("connection_data")

        self._client_config = {
            "uri": self._connection_data.get("persist_directory"),
            "api_key": self._connection_data.get("api_key", None),
            "region": self._connection_data.get("region"),
            "host_override": self._connection_data.get("host_override"),
        }

        # uri is required either for LanceDB Cloud or local
        if not self._client_config["uri"]:
            raise Exception(
                "persist_directory is required for LanceDB connection!"
            )
        # uri, api_key and region is required either for LanceDB Cloud
        elif self._client_config["uri"] and self._client_config["api_key"] and not self._client_config["region"]:
            raise Exception(
                "region is required for LanceDB Cloud connection!"
            )

        self._client = None
        self.is_connected = False
        self.connect()

    def _get_client(self):
        client_config = self._client_config
        if client_config is None:
            raise Exception("Client config is not set!")
        return lancedb.connect(**client_config)

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """Connect to a LanceDB database."""
        if self.is_connected is True:
            return
        try:
            self._client = self._get_client()
            self.is_connected = True
        except Exception as e:
            logger.error(f"Error connecting to LanceDB client, {e}!")
            self.is_connected = False

    def disconnect(self):
        """Close the database connection."""
        if self.is_connected is False:
            return
        self._client = None
        self.is_connected = False

    def check_connection(self):
        """Check the connection to the LanceDB database."""
        response_code = StatusResponse(False)
        need_to_close = self.is_connected is False
        try:
            self._client.table_names()
            response_code.success = True
        except Exception as e:
            logger.error(f"Error connecting to LanceDB , {e}!")
            response_code.error_message = str(e)
        finally:
            if response_code.success is True and need_to_close:
                self.disconnect()
            if response_code.success is False and self.is_connected is True:
                self.is_connected = False

        return response_code

    def _get_lancedb_operator(self, operator: FilterOperator) -> str:
        # The in values are not returned with () and only one element is returned. Bug
        mapping = {
            FilterOperator.EQUAL: "=",
            FilterOperator.NOT_EQUAL: "!=",
            FilterOperator.LESS_THAN: "<",
            FilterOperator.LESS_THAN_OR_EQUAL: "<=",
            FilterOperator.GREATER_THAN: ">",
            FilterOperator.GREATER_THAN_OR_EQUAL: ">=",
            # FilterOperator.IN: "in",
            # FilterOperator.NOT_IN: "not in",
            FilterOperator.LIKE: "like",
            FilterOperator.NOT_LIKE: "not like",
            FilterOperator.IS_NULL: "is null",
            FilterOperator.IS_NOT_NULL: "is not null",
        }

        if operator not in mapping:
            raise Exception(f"Operator {operator} is not supported by LanceDB!")

        return mapping[operator]

    def _translate_condition(
        self, conditions: List[FilterCondition]
    ) -> Optional[dict]:
        """
        Translate a list of FilterCondition objects to string that can be used by LanceDB.
        E.g.,
        [
            FilterCondition(
                column="content",
                op=FilterOperator.NOT_EQUAL,
                value="a",
            ),
            FilterCondition(
                column="id",
                op=FilterOperator.EQUAL,
                value="6",
            )
        ]
        -->
            "content != 'a' and id = '6'"
        """
        # we ignore all non-metadata conditions
        if not conditions:
            return
        filtered_conditions = [
            condition
            for condition in conditions
            if condition.column.startswith(TableField.ID.value) or condition.column.startswith(TableField.CONTENT.value)
        ]

        if len(filtered_conditions) == 0:
            return None

        # generate the LanceDB filter string
        lancedb_conditions = []
        for condition in filtered_conditions:
            if isinstance(condition.value, str):
                condition.value = f"'{condition.value}'"
            condition_key = condition.column.split(".")[-1]
            lancedb_conditions.append(
                ' '.join([condition_key, self._get_lancedb_operator(condition.op), str(condition.value)])
            )
        # Combine all conditions into a single string and return
        return " and ".join(lancedb_conditions) if lancedb_conditions else None

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> HandlerResponse:
        try:
            # Load collection table
            collection = self._client.open_table(table_name)
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error loading collection {table_name}: {e}",
            )
        filters = self._translate_condition(conditions)
        # check if embedding vector filter is present
        vector_filter = (
            []
            if conditions is None
            else [
                condition
                for condition in conditions
                if condition.column == TableField.SEARCH_VECTOR.value
            ]
        )

        if len(vector_filter) > 0:
            vector_filter = vector_filter[0]
        else:
            vector_filter = None

        if vector_filter is not None:
            vec = json.loads(vector_filter.value) if isinstance(vector_filter.value, str) else vector_filter.value
            result = collection.search(vec).select(columns).to_pandas()
            result = result.rename(columns={"_distance": TableField.DISTANCE.value})
        else:
            result = self._client.open_table(table_name).to_pandas()

        new_columns = columns + [TableField.DISTANCE.value] if TableField.DISTANCE.value in result.columns else columns

        col_str = ', '.join([col for col in new_columns if col in (TableField.ID.value, TableField.CONTENT.value, TableField.METADATA.value, TableField.EMBEDDINGS.value, TableField.DISTANCE.value)])

        where_str = f'where {filters}' if filters else ''
        # implementing limit and offset. Not supported natively in lancedb
        if limit and offset:
            sql = f"""select {col_str} from result {where_str} limit {limit} offset {offset}"""
        elif limit and not offset:
            sql = f"""select {col_str} from result {where_str} limit {limit}"""
        elif offset and not limit:
            sql = f"""select {col_str} from result {where_str} offset {offset}"""
        else:
            sql = f"""select {col_str} from result {where_str}"""

        data_df = duckdb.query(sql).to_df()
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=data_df)

    def insert(
        self, table_name: str, data: pd.DataFrame, columns: List[str] = None
    ) -> HandlerResponse:
        """
        Insert data into the LanceDB database.
        In case of create table statements the there is a mismatch between the column types of the `data` pandas dataframe filled with data
        and the empty base table column types which raises a pa.lib.ArrowNotImplementedError, in that case the base table is deleted (doesn't matter as it is empty)
        and recreated with the right datatypes
        """
        try:
            collection = self._client.open_table(table_name)
            df = data[[TableField.ID.value, TableField.CONTENT.value, TableField.METADATA.value, TableField.EMBEDDINGS.value]]
            pa_data = pa.Table.from_pandas(df, preserve_index=False)
            vec_data = vec_to_table(df[TableField.EMBEDDINGS.value].values.tolist())
            new_pa_data = pa_data.append_column("vector", vec_data["vector"])
            collection.add(new_pa_data)
        except pa.lib.ArrowNotImplementedError:
            collection_df = collection.to_pandas()
            column_dtypes = collection_df.dtypes
            df = df.astype(column_dtypes)
            new_df = pd.concat([collection_df, df])
            new_df['id'] = new_df['id'].apply(str)
            pa_data = pa.Table.from_pandas(new_df, preserve_index=False)
            vec_data = vec_to_table(df[TableField.EMBEDDINGS.value].values.tolist())
            new_pa_data = pa_data.append_column("vector", vec_data["vector"])
            self.drop_table(table_name)
            self._client.create_table(table_name, new_pa_data)
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Unable to insert data into collection `{table_name}`: {e}"
            )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def update(
        self, table_name: str, data: pd.DataFrame, columns: List[str] = None
    ) -> HandlerResponse:
        """
        Update data in the LanceDB database.
        TODO: not implemented yet
        """
        return super().update(table_name, data, columns)

    def delete(
        self, table_name: str, conditions: List[FilterCondition] = None
    ) -> HandlerResponse:
        try:
            filters = self._translate_condition(conditions)
            if filters is None:
                raise Exception("Delete query must have at least one condition!")
            collection = self._client.open_table(table_name)
            collection.delete(filters)
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error deleting from collection `{table_name}`: {e}",
            )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def create_table(self, table_name: str, if_not_exists=True) -> HandlerResponse:
        """
        Create a collection with the given name in the LanceDB database.
        """
        try:
            data = {
                TableField.ID.value: str,
                TableField.CONTENT.value: str,
                TableField.METADATA.value: object,
                TableField.EMBEDDINGS.value: object,
            }
            df = pd.DataFrame(columns=data.keys()).astype(data)
            self._client.create_table(table_name, df)
        except Exception as e:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Unable to create collection `{table_name}`: {e}"
            )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def drop_table(self, table_name: str, if_exists=True) -> HandlerResponse:
        """
        Delete a collection from the LanceDB database.
        """
        try:
            self._client.drop_table(table_name)
        except ValueError:
            if if_exists:
                return Response(resp_type=RESPONSE_TYPE.OK)
            else:
                return Response(
                    resp_type=RESPONSE_TYPE.ERROR,
                    error_message=f"Table {table_name} does not exist!",
                )
        return Response(resp_type=RESPONSE_TYPE.OK)

    def get_tables(self) -> HandlerResponse:
        """
        Get the list of collections in the LanceDB database.
        """
        collections = self._client.table_names()
        collections_name = pd.DataFrame(
            columns=["table_name"],
            data=collections,
        )
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=collections_name)

    def get_columns(self, table_name: str) -> HandlerResponse:
        # check if collection exists
        try:
            df = self._client.open_table(table_name).to_pandas()
            column_df = pd.DataFrame(df.dtypes).reset_index()
            column_df.columns = ['column_name', 'data_type']
        except ValueError:
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Table {table_name} does not exist!",
            )
        return Response(resp_type=RESPONSE_TYPE.TABLE, data_frame=column_df)


connection_args = OrderedDict(
    persist_directory={
        "type": ARG_TYPE.STR,
        "description": "The uri of the database.",
        "required": True,
    },
    api_key={
        "type": ARG_TYPE.STR,
        "description": "If presented, connect to LanceDB cloud. Otherwise, connect to a database on file system or cloud storage.",
        "required": False,
    },
    region={
        "type": ARG_TYPE.STR,
        "description": "The region to use for LanceDB Cloud.",
        "required": False,
    },
    host_override={
        "type": ARG_TYPE.STR,
        "description": "The override url for LanceDB Cloud.",
        "required": False,
    },
)

connection_args_example = OrderedDict(
    persist_directory="~/lancedb",
    api_key=None,
    region="us-west-2",
    host_override=None,
)



from typing import Dict, Optional

import dill
import pandas as pd
from dataprep_ml.recommenders import RecommenderPreprocessor
from lightfm import LightFM
from lightfm.cross_validation import random_train_test_split
from lightfm.evaluation import auc_score, precision_at_k, recall_at_k

from mindsdb.integrations.handlers.lightfm_handler.helpers import (
    get_item_item_recommendations,
    get_user_item_recommendations,
)
from mindsdb.integrations.handlers.lightfm_handler.settings import ModelParameters
from mindsdb.integrations.libs.base import BaseMLEngine


class LightFMHandler(BaseMLEngine):
    """
    Integration with the lightfm Recommender library.
    """

    name = "lightfm"

    # todo add ability to partially update model based on new data for existing users, items
    # todo add support for hybrid recommender
    def create(self, target: str, df: pd.DataFrame = None, args: Optional[Dict] = None):

        args = args["using"]

        # get model parameters if defined by user - else use default values

        user_defined_model_params = list(
            filter(lambda x: x in args, ["learning_rate", "loss", "epochs"])
        )
        args["model_params"] = {
            model_param: args[model_param] for model_param in user_defined_model_params
        }
        model_parameters = ModelParameters(**args["model_params"])

        # store model parameters
        args["model_params"] = model_parameters.dict()

        rec_preprocessor = RecommenderPreprocessor(
            interaction_data=df,
            user_id_column_name=args["user_id"],
            item_id_column_name=args["item_id"],
            threshold=args["threshold"],
        )

        # preprocess data
        preprocessed_data = rec_preprocessor.preprocess()

        args["n_users_items"] = rec_preprocessor.n_users_items

        # get item idx to id and user idx to id maps
        args["item_idx_to_id_map"] = preprocessed_data.idx_item_map
        args["user_idx_to_id_map"] = preprocessed_data.idx_user_map

        random_state = 42

        # run evaluation if specified
        if args.get("evaluation"):

            evaluation_metrics = self.evaluate(
                preprocessed_data.interaction_matrix, random_state, model_parameters
            )
            # convert to float to str so it can be stored in json
            args["evaluation_metrics"] = {
                k: str(v) for k, v in evaluation_metrics.items()
            }

        # train model
        model = LightFM(
            learning_rate=model_parameters.learning_rate,
            loss=model_parameters.loss,
            random_state=random_state,
        )
        model.fit(preprocessed_data.interaction_matrix, epochs=model_parameters.epochs)

        self.model_storage.file_set("model", dill.dumps(model))
        self.model_storage.json_set("args", args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[dict] = None):

        predict_params = args["predict_params"]

        # if user doesn't specify recommender type, default to user_item
        try:
            recommender_type = df["recommender_type"].tolist()[0]
        except KeyError:
            recommender_type = predict_params.get("recommender_type", "user_item")

        args = self.model_storage.json_get("args")

        # get item_id to idx and user_id to idx maps
        args["item_id_to_idx_map"] = dict(
            zip(args["item_idx_to_id_map"].values(), args["item_idx_to_id_map"].keys())
        )
        args["user_id_to_idx_map"] = dict(
            zip(args["user_idx_to_id_map"].values(), args["user_idx_to_id_map"].keys())
        )

        model = dill.loads(self.model_storage.file_get("model"))

        n_users = args["n_users_items"][0]
        n_items = args["n_users_items"][1]
        item_ids, user_ids = None, None

        if recommender_type == "user_item":
            if df is not None:

                if args["item_id"] in df.columns:
                    n_items = df[args["item_id"]].nunique()
                    item_ids = df[args["item_id"]].unique().tolist()

                if args["user_id"] in df.columns:
                    n_users = df[args["user_id"]].nunique()
                    user_ids = df[args["user_id"]].unique().tolist()

            return get_user_item_recommendations(
                n_users=n_users,
                n_items=n_items,
                args=args,
                model=model,
                item_ids=item_ids,
                user_ids=user_ids,
            )

        elif recommender_type == "item_item":
            if df is not None and args["item_id"] in df.columns:
                item_ids = df[args["item_id"]].unique().tolist()

            return get_item_item_recommendations(
                model=model,
                args=args,
                item_ids=item_ids,
            )

        elif recommender_type == "user_user":
            raise NotImplementedError(
                "user_user recommendation type is not implemented yet"
            )

        else:
            raise ValueError(
                "recommender_type must be either 'user_item', 'item_item' or 'user_user'"
            )

    def describe(self, attribute=None):
        model_args = self.model_storage.json_get("args")

        if attribute == "model":
            return pd.DataFrame({k: [model_args[k]] for k in ["model_params"]})

        elif attribute == "features":
            return pd.DataFrame(
                {
                    "n_users_items": [model_args["n_users_items"]],
                }
            )

        elif attribute == "info":

            model_metrics = model_args["evaluation_metrics"]

            info_dict = {k: [model_metrics[k]] for k in ["auc", "precision", "recall"]}
            info_dict["user_id"] = [model_args["user_id"]]
            info_dict["item_id"] = [model_args["item_id"]]

            return pd.DataFrame(info_dict)

        else:
            tables = ["info", "features", "model"]
            return pd.DataFrame(tables, columns=["tables"])

    def evaluate(self, interaction_matrix, random_state, model_parameters):

        train, test = random_train_test_split(
            interaction_matrix,
            test_percentage=0.2,
            random_state=random_state,
        )
        model = LightFM(
            learning_rate=model_parameters.learning_rate,
            loss=model_parameters.loss,
            random_state=random_state,
        )

        model.fit(train, epochs=model_parameters.epochs)

        evaluation_metrics = dict(
            auc=auc_score(model, test, train).mean(),
            precision=precision_at_k(model, test, train).mean(),
            recall=recall_at_k(model, test, train).mean(),
        )

        return evaluation_metrics



import json
from collections import defaultdict
from typing import List

from mindsdb.integrations.handlers.rag_handler.settings import (
    LLMLoader,
    PersistedVectorStoreLoader,
    PersistedVectorStoreLoaderConfig,
    RAGBaseParameters,
    RAGHandlerParameters,
    load_embeddings_model,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class RAGQuestionAnswerer:
    """A class for using a RAG model for question answering"""

    def __init__(self, args: RAGBaseParameters):

        self.output_data = defaultdict(list)

        self.args = args

        self.embeddings_model = args.embeddings_model
        if self.embeddings_model is None:
            self.embeddings_model = load_embeddings_model(args.embeddings_model_name)

        self.persist_directory = args.vector_store_storage_path

        self.collection_name = args.collection_name

        vector_store_config = PersistedVectorStoreLoaderConfig(
            vector_store_name=args.vector_store_name,
            embeddings_model=self.embeddings_model,
            persist_directory=self.persist_directory,
            collection_name=self.collection_name,
        )

        self.vector_store_loader = PersistedVectorStoreLoader(vector_store_config)

        self.persisted_vector_store = self.vector_store_loader.load_vector_store()

        self.prompt_template = args.prompt_template

        if isinstance(args, RAGHandlerParameters):

            llm_config = {"llm_config": args.llm_params.dict()}

            llm_loader = LLMLoader(**llm_config)

            self.llm = llm_loader.load_llm()

    def __call__(self, question: str) -> defaultdict:
        return self.query(question)

    def _prepare_prompt(self, vector_store_response, question) -> str:

        context = [doc.page_content for doc in vector_store_response]

        combined_context = "\n\n".join(context)

        if self.args.summarize_context:
            return self.summarize_context(combined_context, question)

        return self.prompt_template.format(question=question, context=combined_context)

    def summarize_context(self, combined_context: str, question: str) -> str:

        summarization_prompt_template = self.args.summarization_prompt_template

        summarization_prompt = summarization_prompt_template.format(
            context=combined_context, question=question
        )

        summarized_context = self.llm(prompt=summarization_prompt)

        return self.prompt_template.format(
            question=question, context=self.extract_generated_text(summarized_context)
        )

    def query_vector_store(self, question: str) -> List:

        return self.persisted_vector_store.similarity_search(
            query=question,
            k=self.args.top_k,
        )

    @staticmethod
    def extract_generated_text(response: str) -> str:
        """Extract generated text from LLM response"""

        if isinstance(response, str):
            data = json.loads(response)
        else:
            data = response

        try:
            if "choices" in data:
                return data["choices"][0]["text"]
            else:
                logger.info(
                    f"Error extracting generated text: failed to parse response {response}"
                )
                return response

        except Exception as e:
            raise Exception(
                f"{e} Error extracting generated text: failed to parse response {response}"
            )

    def query(self, question: str) -> defaultdict:
        """Post process LLM response"""
        llm_response, vector_store_response = self._query(question)

        result = defaultdict(list)
        extracted_text = self.extract_generated_text(llm_response)
        result["answer"].append(extracted_text)

        sources = defaultdict(list)

        for idx, document in enumerate(vector_store_response):
            sources["sources_content"].append(document.page_content)
            sources["sources_document"].append(document.metadata.get("source", None))
            sources["column"].append(document.metadata.get("column", None))
            sources["sources_row"].append(document.metadata.get("row", None))

        result["source_documents"].append(dict(sources))

        return result

    def _query(self, question: str):
        logger.debug(f"Querying: {question}")

        vector_store_response = self.query_vector_store(question)

        formatted_prompt = self._prepare_prompt(vector_store_response, question)

        llm_response = self.llm(prompt=formatted_prompt)

        return llm_response, vector_store_response



from mindsdb_sql.parser import ast

from mindsdb.integrations.handlers.utilities.query_utilities.base_query_utilities import BaseQueryParser
from mindsdb.integrations.handlers.utilities.query_utilities.base_query_utilities import BaseQueryExecutor


class DELETEQueryParser(BaseQueryParser):
    """
    Parses a DELETE query into its component parts.

    Parameters
    ----------
    query : ast.Delete
        Given SQL DELETE query.
    """
    def __init__(self, query: ast.Delete):
        super().__init__(query)
    
    def parse_query(self):
        """
        Parses a SQL DELETE statement into its components: WHERE.
        """
        where_conditions = self.parse_where_clause()

        return where_conditions
    

class DELETEQueryExecutor(BaseQueryExecutor):
    """
    Executes a DELETE query.

    Parameters
    ----------
    df : pd.DataFrame
        Given table.
    where_conditions : List[List[Text]]
        WHERE conditions of the query.

    NOTE: This class DOES NOT delete the relevant records of the entity for you, it will simply return the records that need to be deleted based on the WHERE conditions.
    
          This class expects all of the entities to be passed in as a DataFrane and filters out the relevant records based on the WHERE conditions.
          Because all of the records need to be extracted to be passed in as a DataFrame, this class is not very computationally efficient.
          Therefore, DO NOT use this class if the API/SDK that you are using supports deleting records in bulk.
    """



from collections import OrderedDict
from typing import List, Union

import pandas as pd
import psycopg
from mindsdb_sql import ASTNode, Parameter, Identifier, Update, BinaryOperation
from pgvector.psycopg import register_vector

from mindsdb.integrations.handlers.postgres_handler.postgres_handler import (
    PostgresHandler,
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.response import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    VectorStoreHandler,
)
from mindsdb.utilities import log
from mindsdb.utilities.profiler import profiler

logger = log.getLogger(__name__)


# todo Issue #7316 add support for different indexes and search algorithms e.g. cosine similarity or L2 norm
class PgVectorHandler(VectorStoreHandler, PostgresHandler):
    """This handler handles connection and execution of the PostgreSQL with pgvector extension statements."""

    name = "pgvector"

    def __init__(self, name: str, **kwargs):

        super().__init__(name=name, **kwargs)
        self.connect()

    @profiler.profile()
    def connect(self) -> psycopg.connection:
        """
        Handles the connection to a PostgreSQL database instance.
        """
        self.connection = super().connect()

        with self.connection.cursor() as cur:
            try:
                # load pg_vector extension
                cur.execute("CREATE EXTENSION IF NOT EXISTS vector")
                logger.info("pg_vector extension loaded")

            except psycopg.Error as e:
                logger.error(
                    f"Error loading pg_vector extension, ensure you have installed it before running, {e}!"
                )
                return HandlerResponse(resp_type=RESPONSE_TYPE.ERROR, error_message=str(e))

        # register vector type with psycopg2 connection
        register_vector(self.connection)

        return self.connection

    @staticmethod
    def _translate_conditions(conditions: List[FilterCondition]) -> Union[dict, None]:
        """
        Translate filter conditions to a dictionary
        """

        if conditions is None:
            return {}

        return {
            condition.column.split(".")[-1]: {
                "op": condition.op.value,
                "value": condition.value,
            }
            for condition in conditions
        }

    @staticmethod
    def _construct_where_clause(filter_conditions=None):
        """
        Construct where clauses from filter conditions
        """
        if filter_conditions is None:
            return ""

        where_clauses = []

        for key, value in filter_conditions.items():
            if key == "embeddings":
                continue
            if value['op'].lower() == 'in':
                values = list(repr(i) for i in value['value'])
                value['value'] = '({})'.format(', '.join(values))
            else:
                value['value'] = repr(value['value'])
            where_clauses.append(f'{key} {value["op"]} {value["value"]}')

        if len(where_clauses) > 1:
            return f"WHERE{' AND '.join(where_clauses)}"
        elif len(where_clauses) == 1:
            return f"WHERE {where_clauses[0]}"
        else:
            return ""

    @staticmethod
    def _construct_full_after_from_clause(
        offset_clause: str,
        limit_clause: str,
        where_clause: str,
    ) -> str:

        return f"{where_clause} {offset_clause} {limit_clause}"

    def _build_select_query(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        offset: int = None,
    ) -> str:
        """
        given inputs, build string query
        """
        limit_clause = f"LIMIT {limit}" if limit else ""
        offset_clause = f"OFFSET {offset}" if offset else ""

        # translate filter conditions to dictionary
        filter_conditions = self._translate_conditions(conditions)

        # check if search vector is in filter conditions
        embedding_search = filter_conditions.get("embeddings", None)

        # given filter conditions, construct where clause
        where_clause = self._construct_where_clause(filter_conditions)

        # construct full after from clause, where clause + offset clause + limit clause
        after_from_clause = self._construct_full_after_from_clause(
            where_clause, offset_clause, limit_clause
        )

        if columns is None:
            targets = '*'
        else:
            targets = ', '.join(columns)


        if filter_conditions:

            if embedding_search:
                # if search vector, return similar rows, apply other filters after if any
                search_vector = filter_conditions["embeddings"]["value"][0]
                filter_conditions.pop("embeddings")
                return f"SELECT {targets} FROM {table_name} ORDER BY embeddings <=> '{search_vector}' {after_from_clause}"
            else:
                # if filter conditions, return filtered rows
                return f"SELECT {targets} FROM {table_name} {after_from_clause}"
        else:
            # if no filter conditions, return all rows
            return f"SELECT {targets} FROM {table_name} {after_from_clause}"

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> pd.DataFrame:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        if columns is None:
            columns = ["id", "content", "embeddings", "metadata"]

        with self.connection.cursor() as cur:
            query = self._build_select_query(table_name, columns, conditions, limit, offset)
            cur.execute(query)

            self.connection.commit()
            result = cur.fetchall()

        result = pd.DataFrame(result, columns=columns)
        # ensure embeddings are returned as string so they can be parsed by mindsdb
        if "embeddings" in columns:
            result["embeddings"] = result["embeddings"].astype(str)

        return result

    def create_table(self, table_name: str, if_not_exists=True):
        """
        Run a create table query on the pgvector database.
        """
        with self.connection.cursor() as cur:
            cur.execute(
                f"CREATE TABLE IF NOT EXISTS {table_name} (id text PRIMARY KEY, content text, embeddings vector, metadata jsonb)"
            )
            self.connection.commit()

    def insert(
        self, table_name: str, data: pd.DataFrame
    ):
        """
        Insert data into the pgvector table database.
        """
        data_dict = data.to_dict(orient="list")
        transposed_data = list(zip(*data_dict.values()))

        columns = ", ".join(data.keys())
        values = ", ".join(["%s"] * len(data.keys()))

        insert_statement = f"INSERT INTO {table_name} ({columns}) VALUES ({values})"

        with self.connection.cursor() as cur:
            cur.executemany(insert_statement, transposed_data)
            self.connection.commit()

    def update(
        self, table_name: str, data: pd.DataFrame, key_columns: List[str] = None
    ):
        """
        Udate data into the pgvector table database.
        """

        where = None
        update_columns = {}

        for col in data.columns:
            value = Parameter('%s')

            if col in key_columns:
                cond = BinaryOperation(
                    op='=',
                    args=[Identifier(col), value]
                )
                if where is None:
                    where = cond
                else:
                    where = BinaryOperation(
                        op='AND',
                        args=[where, cond]
                    )
            else:
                update_columns[col] = value

        query = Update(
            table=Identifier(table_name),
            update_columns=update_columns,
            where=where
        )

        with self.connection.cursor() as cur:
            transposed_data = []
            for _, record in data.iterrows():
                row = [
                    record[col]
                    for col in update_columns.keys()
                ]
                for key_column in key_columns:
                    row.append(record[key_column])
                transposed_data.append(row)

            query_str = self.renderer.get_string(query)
            cur.executemany(query_str, transposed_data)
            self.connection.commit()

    def delete(
        self, table_name: str, conditions: List[FilterCondition] = None
    ):

        filter_conditions = self._translate_conditions(conditions)
        where_clause = self._construct_where_clause(filter_conditions)

        with self.connection.cursor() as cur:

            # convert search embedding to string

            # we need to use the <-> operator to search for similar vectors,
            # so we need to convert the string to a vector and also use a threshold (e.g. 0.5)

            query = (
                f"DELETE FROM {table_name} {where_clause}"
            )
            cur.execute(query)
            self.connection.commit()

    def drop_table(self, table_name: str, if_exists=True):
        """
        Run a drop table query on the pgvector database.
        """
        with self.connection.cursor() as cur:
            cur.execute(f"DROP TABLE IF EXISTS {table_name}")
            self.connection.commit()


connection_args = OrderedDict(
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate with the PostgreSQL server.",
        "required": True,
        "label": "User",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password to authenticate the user with the PostgreSQL server.",
        "required": True,
        "label": "Password",
    },
    database={
        "type": ARG_TYPE.STR,
        "description": "The database name to use when connecting with the PostgreSQL server.",
        "required": True,
        "label": "Database",
    },
    host={
        "type": ARG_TYPE.STR,
        "description": "The host name or IP address of the PostgreSQL server. NOTE: use '127.0.0.1' instead of 'localhost' to connect to local server.",
        "required": True,
        "label": "Host",
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "The TCP/IP port of the PostgreSQL server. Must be an integer.",
        "required": True,
        "label": "Port",
    },
    schema={
        "type": ARG_TYPE.STR,
        "description": "The schema in which objects are searched first.",
        "required": False,
        "label": "Schema",
    },
    sslmode={
        "type": ARG_TYPE.STR,
        "description": "sslmode that will be used for connection.",
        "required": False,
        "label": "sslmode",
    },
)

connection_args_example = OrderedDict(
    host="127.0.0.1", port=5432, user="root", password="password", database="database"
)



import github
from collections import OrderedDict

from mindsdb.integrations.handlers.github_handler.github_tables import (
    GithubIssuesTable,
    GithubPullRequestsTable,
    GithubCommitsTable,
    GithubReleasesTable,
    GithubBranchesTable,
    GithubContributorsTable,
    GithubMilestonesTable,
    GithubProjectsTable
)

from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from mindsdb.utilities import log
from mindsdb_sql import parse_sql


logger = log.getLogger(__name__)

class GithubHandler(APIHandler):
    """The GitHub handler implementation"""

    def __init__(self, name: str, **kwargs):
        """Initialize the GitHub handler.

        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.repository = connection_data["repository"]
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        github_issues_data = GithubIssuesTable(self)
        github_pull_requests_data = GithubPullRequestsTable(self)
        github_commits_data = GithubCommitsTable(self)
        github_releases_data = GithubReleasesTable(self)
        github_branches_data = GithubBranchesTable(self)
        github_contributors_data = GithubContributorsTable(self)
        github_milestones_data = GithubMilestonesTable(self)
        github_projects_data = GithubProjectsTable(self)

        self._register_table("issues", github_issues_data)
        self._register_table("pull_requests", github_pull_requests_data)
        self._register_table("commits", github_commits_data)
        self._register_table("releases", github_releases_data)
        self._register_table("branches", github_branches_data)
        self._register_table("contributors", github_contributors_data)
        self._register_table("milestones", github_milestones_data)
        self._register_table("projects", github_projects_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.

        Returns
        -------
        StatusResponse
            connection object
        """

        connection_kwargs = {}

        if self.connection_data.get("api_key", None):
            connection_kwargs["login_or_token"] = self.connection_data["api_key"]

        if self.connection_data.get("github_url", None):
            connection_kwargs["base_url"] = self.connection_data["github_url"]

        self.connection = github.Github(**connection_kwargs)
        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)

        try:
            self.connect()
            if self.connection_data.get("api_key", None):
                current_user = self.connection.get_user().name
                logger.info(f"Authenticated as user {current_user}")
            else:
                logger.info("Proceeding without an API key")

            current_limit = self.connection.get_rate_limit()
            logger.info(f"Current rate limit: {current_limit}")
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to GitHub API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


connection_args = OrderedDict(
    repository={
        "type": ARG_TYPE.STR,
        "description": " GitHub repository name.",
        "required": True,
        "label": "Repository",
    },
    api_key={
        "type": ARG_TYPE.PWD,
        "description": "Optional GitHub API key to use for authentication.",
        "required": False,
        "label": "Api key",
    },
    github_url={
        "type": ARG_TYPE.STR,
        "description": "Optional GitHub URL to connect to a GitHub Enterprise instance.",
        "required": False,
        "label": "Github url",
    },
)

connection_args_example = OrderedDict(
    repository="mindsdb/mindsdb", 
    api_key="ghp_z91InCQZWZAMlddOzFCX7xHJrf9Fai35HT7", 
    github_url="https://github.com/mindsdb/mindsdb"
)



__title__ = 'MindsDB CoinBase handler'
__package_name__ = 'mindsdb_coinbase_handler'
__version__ = '0.0.1'
__description__ = "MindsDB handler for the CoinBase API"
__author__ = 'Bustedware LLC'
__github__ = 'https://github.com/bustedware'
__pypi__ = 'https://pypi.org/project/mindsdb/'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023 - mindsdb'



from typing import Optional, Dict
import os

import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine
from pycaret.classification import ClassificationExperiment
from pycaret.regression import RegressionExperiment
from pycaret.time_series import TSForecastingExperiment
from pycaret.clustering import ClusteringExperiment
from pycaret.anomaly import AnomalyExperiment


class PyCaretHandler(BaseMLEngine):
    name = 'pycaret'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """Create and train model on given data"""
        # parse args
        if 'using' not in args:
            raise Exception("PyCaret engine requires a USING clause! Refer to its documentation for more details.")
        using = args['using']
        if df is None:
            raise Exception("PyCaret engine requires a some data to initialize!")
        # create experiment
        s = self._get_experiment(using['model_type'])
        s.setup(df, **self._get_experiment_setup_kwargs(using, args['target']))
        # train model
        model = self._train_model(s, using)
        # save model and args
        model_file_path = os.path.join(self.model_storage.fileStorage.folder_path, 'model')
        s.save_model(model, model_file_path)
        self.model_storage.json_set('saved_args', {
            **args['using'],
            'model_path': model_file_path
        })

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> pd.DataFrame:
        """Predict on the given data"""
        # load model
        saved_args = self.model_storage.json_get('saved_args')
        s = self._get_experiment(saved_args['model_type'])
        model = s.load_model(saved_args['model_path'])
        # predict and return
        return self._predict_model(s, model, df, saved_args)

    def _get_experiment(self, model_type):
        """Returns one of the types of experiments in PyCaret"""
        if model_type == "classification":
            return ClassificationExperiment()
        elif model_type == "regression":
            return RegressionExperiment()
        elif model_type == "time_series":
            return TSForecastingExperiment()
        elif model_type == "clustering":
            return ClusteringExperiment()
        elif model_type == "anomaly":
            return AnomalyExperiment()
        else:
            raise Exception(f"Unrecognized model type '{model_type}'")

    def _get_experiment_setup_kwargs(self, args: Dict, target: str):
        """Returns the arguments that need to passed in setup function for the experiment"""
        model_type = args['model_type']
        # copy setup kwargs
        kwargs = self._select_keys(args, "setup_")
        # return kwargs
        if model_type == 'classification' or model_type == 'regression' or model_type == 'time_series':
            return {**kwargs, 'target': target}
        elif model_type == 'clustering' or model_type == 'anomaly':
            return {**kwargs}
        raise Exception(f"Unrecognized model type '{model_type}'")

    def _predict_model(self, s, model, df, args):
        """Apply predictor arguments and get predictions"""
        model_type = args["model_type"]
        kwargs = self._select_keys(args, "predict_")
        if (
            model_type == 'classification'
            or model_type == 'regression'
            or model_type == 'clustering'
            or model_type == 'anomaly'
        ):
            kwargs["data"] = df
        elif model_type == 'time_series':
            # do nothing
            pass
        else:
            raise Exception(f"Unrecognized model type '{model_type}'")
        return s.predict_model(model, **kwargs)

    def _train_model(self, experiment, args):
        """Train the model and return the best (if applicable)"""
        model_type = args['model_type']
        model_name = args['model_name']
        kwargs = self._select_keys(args, "create_")
        if (
            model_type == 'classification'
            or model_type == 'regression'
            or model_type == 'time_series'
        ) and model_name == 'best':
            return experiment.compare_models(**kwargs)
        if model_name == 'best':
            raise Exception("Specific model name must be provided for clustering or anomaly tasks")
        return experiment.create_model(model_name, **kwargs)

    def _select_keys(self, d, prefix):
        """Selects keys with given prefix and returns a new dict"""
        result = {}
        for k in d:
            if k.startswith(prefix):
                result[k[len(prefix):]] = d[k]
        return result



import os
from google.auth.transport.requests import Request
from google.oauth2 import service_account
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build
import pandas as pd
from pandas import DataFrame
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from .google_books_tables import BookshelvesTable, VolumesTable
from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log

logger = log.getLogger(__name__)

class GoogleBooksHandler(APIHandler):
    """
        A class for handling connections and interactions with the Google Books API.
    """
    name = 'google_books'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the Google Books API handler.
        Args:
            name (str): name of the handler
            kwargs (dict): additional arguments
        """
        super().__init__(name)
        self.token = None
        self.service = None
        self.connection_data = kwargs.get('connection_data', {})
        self.credentials_file = self.connection_data.get('credentials', None)
        self.credentials = None
        self.scopes = ['https://www.googleapis.com/auth/books']
        self.is_connected = False
        self.connection = None
        bookshelves = BookshelvesTable(self)
        self.bookshelves = bookshelves
        self._register_table('bookshelves', bookshelves)
        volumes = VolumesTable(self)
        self.volumes = volumes
        self._register_table('volumes', volumes)

    def connect(self):
        """
        Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.service
        if self.credentials_file:
            if os.path.exists('token_books.json'):
                self.credentials = Credentials.from_authorized_user_file('token_books.json', self.scopes)
            if not self.credentials or not self.credentials.valid:
                if self.credentials and self.credentials.expired and self.credentials.refresh_token:
                    self.credentials.refresh(Request())
                else:
                    self.credentials = service_account.Credentials.from_service_account_file(
                        self.credentials_file, scopes=self.scopes)
            # Save the credentials for the next run
            with open('token_books.json', 'w') as token:
                token.write(self.credentials.to_json())
            self.service = build('books', 'v1', credentials=self.credentials)
        return self.service

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            service = self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Google Books API: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str = None) -> Response:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, api's json etc)
        Returns:
            HandlerResponse
        """
        method_name, params = FuncParser().from_string(query)

        df = self.call_application_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def get_bookshelves(self, params: dict = None) -> DataFrame:
        """
        Get bookshelf from Google Books API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        minShelf = None
        maxShelf = None
        if params['shelf']:
            shelf = int(params['shelf'])
            if params['source']:
                df = service.mylibrary().bookshelves().get(shelf=shelf, userid=params['userid'],
                                                           source=params['source']).execute()
            else:
                df = service.mylibrary().bookshelves().get(shelf=shelf, userid=params['userid']).execute()

            df = pd.DataFrame(df, columns=self.bookshelves.get_columns())
            return df
        elif not params['minShelf'] and params['maxShelf']:
            minShelf = int(params['maxShelf']) - 10
            maxShelf = int(params['maxShelf'])
        elif not params['maxShelf'] and params['minShelf']:
            minShelf = int(params['minShelf'])
            maxShelf = int(params['minShelf']) + 10
        elif params['maxShelf'] and params['minShelf']:
            minShelf = int(params['minShelf'])
            maxShelf = int(params['maxShelf'])

        args = {
            key: value for key, value in params.items() if key not in ['minShelf', 'maxShelf'] and value is not None
        }
        bookshelves = service.bookshelves().list(userid=params['userid'], **args).execute()

        df = pd.DataFrame(bookshelves['items'], columns=self.bookshelves.get_columns())
        if minShelf is not None or maxShelf is not None:
            # Drop bookshelves that are not in the id range
            df = df.drop(df[(df['id'] < minShelf) | (df['id'] > maxShelf)].index)

        return df

    def get_volumes(self, params: dict = None) -> DataFrame:
        """
        Get volumes from Google Books API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        args = {
            key: value for key, value in params.items() if value is not None
        }
        volumes = service.volumes().list(**args).execute()
        df = pd.DataFrame(volumes['items'], columns=self.volumes.get_columns())
        return df

    def call_application_api(self, method_name: str = None, params: dict = None) -> DataFrame:
        """
        Call Google Books API and map the data to pandas DataFrame
        Args:
            method_name (str): method name
            params (dict): query parameters
        Returns:
            DataFrame
        """
        if method_name == 'get_bookshelves':
            return self.get_bookshelves(params)
        elif method_name == 'get_volumes':
            return self.get_volumes(params)
        else:
            raise NotImplementedError(f'Unknown method {method_name}')



import pandas as pd
from typing import Text, List, Dict

from mindsdb_sql.parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions

from mindsdb.integrations.handlers.utilities.query_utilities import SELECTQueryParser, SELECTQueryExecutor


class OpenStreetMapNodeTable(APITable):
    """The OpenStreetMap Nodes Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the OpenStreetMap API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            OpenStreetMap data matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        where_conditions = extract_comparison_conditions(query.where)

        if query.limit:
            result_limit = query.limit.value
        else:
            result_limit = 20

        nodes_df = pd.json_normalize(self.get_nodes(where_conditions=where_conditions, limit=result_limit))

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = nodes_df.columns
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        order_by_conditions = {}
        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[0] == 'nodes':
                    if an_order.field.parts[1] in nodes_df.columns:
                        order_by_conditions["columns"].append(an_order.field.parts[1])

                        if an_order.direction == "ASC":
                            order_by_conditions["ascending"].append(True)
                        else:
                            order_by_conditions["ascending"].append(False)
                    else:
                        raise ValueError(
                            f"Order by unknown column {an_order.field.parts[1]}"
                        )

        select_statement_executor = SELECTQueryExecutor(
            nodes_df,
            selected_columns,
            [],
            order_by_conditions
        )
        nodes_df = select_statement_executor.execute_query()

        return nodes_df

    def get_nodes(self, **kwargs) -> List[Dict]:
        where_conditions = kwargs.get('where_conditions', None)

        area, tags = None, {}
        min_lat, min_lon, max_lat, max_lon = None, None, None, None
        if where_conditions:
            for condition in where_conditions:
                if condition[1] == 'area':
                    area = condition[2]

                elif condition[1] == 'min_lat':
                    min_lat = condition[2]

                elif condition[1] == 'min_lon':
                    min_lon = condition[2]

                elif condition[1] == 'max_lat':
                    max_lat = condition[2]

                elif condition[1] == 'max_lon':
                    max_lon = condition[2]

                else:
                    tags[condition[1]] = condition[2]

        result = self.execute_osm_node_query(
            tags=tags,
            area=area,
            min_lat=min_lat,
            min_lon=min_lon,
            max_lat=max_lat,
            max_lon=max_lon,
            limit=kwargs.get('limit', None)
        )

        nodes = []
        for node in result.nodes:
            node_dict = {
                "id": node.id,
                "lat": node.lat,
                "lon": node.lon,
                "tags": node.tags
            }
            nodes.append(node_dict)
        return nodes

    def execute_osm_node_query(self, tags, area=None, min_lat=None, min_lon=None, max_lat=None, max_lon=None, limit=None):
        query_template = """
        [out:json];
        {area_clause}
        node{area_node_clause}{tags_clause}{bbox};
        out {limit};
        """

        tags_clause = ""
        if tags:
            for tag_key, tag_value in tags.items():
                tags_clause += '["{}"="{}"]'.format(tag_key, tag_value)

        area_clause, area_node_clause = "", ""
        if area:
            area_clause = 'area[name="{}"]->.city;\n'.format(area)
            area_node_clause = "(area.city)"

        bbox_clause = ""
        if min_lat or min_lon or max_lat or max_lon:
            bbox_clause = "{},{},{},{}".format(min_lat, min_lon, max_lat, max_lon)

        limit_clause = limit if limit else ""

        query = query_template.format(
            area_clause=area_clause,
            area_node_clause=area_node_clause,
            tags_clause=tags_clause,
            bbox=bbox_clause,
            limit=limit_clause
        )

        api = self.handler.connect()

        result = api.query(query)
        return result


class OpenStreetMapWayTable(APITable):
    """The OpenStreetMap Ways Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:

        select_statement_parser = SELECTQueryParser(
            query,
            'ways',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        ways_df = pd.json_normalize(self.get_ways(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            ways_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        ways_df = select_statement_executor.execute_query()

        return ways_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_ways(limit=1)).columns.tolist()

    def get_ways(self, **kwargs) -> List[Dict]:

        api_session = self.handler.connect()
        ways = api_session.query("""
            way
            ({{bbox}});
            out;
            """,
                                 # bbox=self.connection_data['bbox']
                                 )
        return [way.to_dict() for way in ways.ways]


class OpenStreetMapRelationTable(APITable):
    """The OpenStreetMap Relations Table implementation"""

    def select_relations(self, query: ast.Select) -> pd.DataFrame:

        select_statement_parser = SELECTQueryParser(
            query,
            'relations',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        relations_df = pd.json_normalize(self.get_relations(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            relations_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        relations_df = select_statement_executor.execute_query()

        return relations_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_relations(limit=1)).columns.tolist()

    def get_relations(self, **kwargs) -> List[Dict]:

        api_session = self.handler.connect()
        relations = api_session.query("""
                    relation
                    ({{bbox}});
                    out;
                    """,
                                      # bbox=self.connection_data['bbox']
                                      )
        return [relation.to_dict() for relation in relations.relations]



import mindsdb.microservices_grpc.ml.common_pb2 as _common_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CreateCall(_message.Message):
    __slots__ = ["args", "context", "df", "target"]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    DF_FIELD_NUMBER: _ClassVar[int]
    TARGET_FIELD_NUMBER: _ClassVar[int]
    args: str
    context: HandlerContextML
    df: bytes
    target: str
    def __init__(self, context: _Optional[_Union[HandlerContextML, _Mapping]] = ..., target: _Optional[str] = ..., df: _Optional[bytes] = ..., args: _Optional[str] = ...) -> None: ...

class HandlerContextML(_message.Message):
    __slots__ = ["context", "handler_params", "integration_id", "predictor_id"]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    HANDLER_PARAMS_FIELD_NUMBER: _ClassVar[int]
    INTEGRATION_ID_FIELD_NUMBER: _ClassVar[int]
    PREDICTOR_ID_FIELD_NUMBER: _ClassVar[int]
    context: str
    handler_params: str
    integration_id: int
    predictor_id: int
    def __init__(self, integration_id: _Optional[int] = ..., predictor_id: _Optional[int] = ..., context: _Optional[str] = ..., handler_params: _Optional[str] = ...) -> None: ...

class PredictCall(_message.Message):
    __slots__ = ["args", "context", "df"]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    DF_FIELD_NUMBER: _ClassVar[int]
    args: str
    context: HandlerContextML
    df: bytes
    def __init__(self, context: _Optional[_Union[HandlerContextML, _Mapping]] = ..., df: _Optional[bytes] = ..., args: _Optional[str] = ...) -> None: ...

class UpdateCall(_message.Message):
    __slots__ = ["args", "context", "df"]
    ARGS_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    DF_FIELD_NUMBER: _ClassVar[int]
    args: str
    context: HandlerContextML
    df: bytes
    def __init__(self, context: _Optional[_Union[HandlerContextML, _Mapping]] = ..., df: _Optional[bytes] = ..., args: _Optional[str] = ...) -> None: ...



from .responder import Responder
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class RespondersCollection():
    def __init__(self):
        self.responders = []

    def find_match(self, query):
        for r in self.responders:
            if r.match(query):
                return r

        msg = f'Is not responder for query: {query}'

        class ErrorResponder(Responder):
            when = {}

            result = {
                "ok": 0.0,
                "errmsg": msg,
                "code": 59,
                "codeName": "CommandNotFound"
            }

        logger.error(msg)
        return ErrorResponder()

    def add(self, when, result):
        self.responders.append(
            Responder(when, result)
        )



import os
from typing import List, Optional
import random
import time
import math

import openai
import tiktoken

import mindsdb.utilities.profiler as profiler
from mindsdb.integrations.handlers.openai_handler.constants import OPENAI_API_BASE


def retry_with_exponential_backoff(
    initial_delay: float = 1,
    hour_budget: float = 0.3,
    jitter: bool = False,
    exponential_base: int = 2,
    errors: tuple = (openai.error.RateLimitError, openai.error.APIConnectionError),
):
    """
    Wrapper to enable optional arguments. It means this decorator always needs to be called with parenthesis:

    > @retry_with_exponential_backoff()  # optional argument override here
    > def f(): [...]

    """  # noqa

    @profiler.profile()
    def _retry_with_exponential_backoff(func):
        """
        Exponential backoff to retry requests on a rate-limited API call, as recommended by OpenAI.
        Loops the call until a successful response or max_retries is hit or an exception is raised.

        Slight changes in the implementation, but originally from:
        https://github.com/openai/openai-cookbook/blob/main/examples/How_to_handle_rate_limits.ipynb
        """  # noqa

        def wrapper(*args, **kwargs):
            num_retries = 0
            delay = initial_delay

            if isinstance(hour_budget, float) or isinstance(hour_budget, int):
                try:
                    max_retries = round(
                        (math.log((hour_budget * 3600) / initial_delay))
                        / math.log(exponential_base)
                    )
                except ValueError:
                    max_retries = 10
            else:
                max_retries = 10
            max_retries = max(1, max_retries)

            while True:
                try:
                    return func(*args, **kwargs)
                except errors as e:
                    if e.error is not None:
                        if (
                            e.error['type'] == 'invalid_request_error'
                            and 'Too many parallel completions' in e.error['message']
                            or 'Please reduce the length of the messages'
                            in e.error['message']
                        ):
                            raise e  # InvalidRequestError triggers batched mode in the previous call
                        if e.error['type'] == 'insufficient_quota':
                            raise Exception(
                                'API key has exceeded its quota, please try 1) increasing it or 2) using another key.'
                            )  # noqa

                    num_retries += 1
                    if num_retries > max_retries:
                        raise Exception(
                            f"Maximum number of retries ({max_retries}) exceeded."
                        )
                    # Increment the delay and wait
                    delay *= exponential_base * (1 + jitter * random.random())
                    time.sleep(delay)

                except openai.error.OpenAIError as e:
                    if e.error is not None and e.error['type'] == 'insufficient_quota':
                        raise Exception(
                            'API key has exceeded its quota, please try 1) increasing it or 2) using another key.'
                        )  # noqa
                    raise e

                except Exception as e:
                    raise e

        return wrapper

    return _retry_with_exponential_backoff


def truncate_msgs_for_token_limit(messages, model_name, max_tokens, truncate='first'):
    """
    Truncates message list to fit within the token limit.
    Note: first message for chat completion models are general directives with the system role, which will ideally be kept at all times.
    """  # noqa
    encoder = tiktoken.encoding_for_model(model_name)
    sys_priming = messages[0:1]
    n_tokens = count_tokens(messages, encoder, model_name)
    while n_tokens > max_tokens:
        if len(messages) == 2:
            return messages[
                :-1
            ]  # edge case: if limit is surpassed by just one input, we remove initial instruction
        elif len(messages) == 1:
            return messages

        if truncate == 'first':
            messages = sys_priming + messages[2:]
        else:
            messages = sys_priming + messages[1:-1]

        n_tokens = count_tokens(messages, encoder, model_name)
    return messages


def count_tokens(messages, encoder, model_name='gpt-3.5-turbo-0301'):
    """Original token count implementation can be found in the OpenAI cookbook."""
    if (
        "gpt-3.5-turbo" in model_name
    ):  # note: future models may deviate from this (only 0301 really complies)
        num_tokens = 0
        for message in messages:
            num_tokens += (
                4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
            )
            for key, value in message.items():
                num_tokens += len(encoder.encode(value))
                if key == "name":  # if there's a name, the role is omitted
                    num_tokens += -1  # role is always required and always 1 token
        num_tokens += 2  # every reply is primed with <im_start>assistant
        return num_tokens
    else:
        raise NotImplementedError(
            f"""_count_tokens() is not presently implemented for model {model_name}."""
        )


def get_available_models(api_key: str) -> List[str]:
    """
    Returns a list of available openai models for the given API key.
    """

    api_base = os.environ.get('OPENAI_API_BASE', OPENAI_API_BASE)
    res = openai.Model.list(api_key=api_key, api_base=api_base)

    return [models["id"] for models in res.data]



import os
from typing import Dict, Optional

import pandas as pd
from anthropic import AI_PROMPT, HUMAN_PROMPT, Anthropic

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log
from mindsdb.utilities.config import Config

logger = log.getLogger(__name__)


class AnthropicHandler(BaseMLEngine):
    """
    Integration with the Anthropic LLM Python Library
    """

    name = "anthropic"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_chat_model = "claude-2"
        self.supported_chat_models = ["claude-1", "claude-2"]
        self.default_max_tokens = 100
        self.generative = True
        self.connection = None

    def create(
        self,
        target: str,
        df: Optional[pd.DataFrame] = None,
        args: Optional[Dict] = None,
    ) -> None:

        if "using" not in args:
            raise Exception(
                "Anthropic engine requires a USING clause! Refer to its documentation for more details."
            )

        if "model" not in args["using"]:
            args["using"]["model"] = self.default_chat_model
        elif args["using"]["model"] not in self.supported_chat_models:
            raise Exception(
                f"Invalid chat model. Please use one of {self.supported_chat_models}"
            )

        if "max_tokens" not in args["using"]:
            args["using"]["max_tokens"] = self.default_max_tokens

        self.model_storage.json_set("args", args)

    def predict(
        self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None
    ) -> None:

        args = self.model_storage.json_get("args")
        api_key = self._get_anthropic_api_key(args)

        self.connection = Anthropic(
            api_key=api_key,
        )

        input_column = args["using"]["column"]

        if input_column not in df.columns:
            raise RuntimeError(f'Column "{input_column}" not found in input data')

        result_df = pd.DataFrame()

        result_df["predictions"] = df[input_column].apply(self.predict_answer)

        result_df = result_df.rename(columns={"predictions": args["target"]})

        return result_df

    def _get_anthropic_api_key(self, args, strict=True):
        """
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. ANTHROPIC_API_KEY env variable
            4. anthropic.api_key setting in config.json
        """

        # 1
        if "api_key" in args["using"]:
            return args["using"]["api_key"]
        # 2
        connection_args = self.engine_storage.get_connection_args()
        if "api_key" in connection_args:
            return connection_args["api_key"]
        # 3
        api_key = os.getenv("ANTHROPIC_API_KEY")
        if api_key is not None:
            return api_key
        # 4
        config = Config()
        anthropic_cfg = config.get("anthropic", {})
        if "api_key" in anthropic_cfg:
            return anthropic_cfg["api_key"]

        if strict:
            raise Exception(
                f'Missing API key "api_key". Either re-create this ML_ENGINE specifying the `api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.'
            )

    def predict_answer(self, text):
        """
        connects with anthropic api to predict the answer for the particular question

        """

        args = self.model_storage.json_get("args")

        completion = self.connection.completions.create(
            model=args["using"]["model"],
            max_tokens_to_sample=args["using"]["max_tokens"],
            prompt=f"{HUMAN_PROMPT} {text} {AI_PROMPT}",
        )

        return completion.completion



from datetime import datetime, timezone
from typing import Text, List, Dict, Any

from mindsdb.integrations.handlers.sharepoint_handler.utils import (
    bearer_token_request,
    get_an_entity,
    delete_an_entity,
    update_an_entity,
    create_an_entity,
)


class SharepointAPI:
    def __init__(
        self, client_id: str = None, client_secret: str = None, tenant_id: str = None
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        self.bearer_token = None
        self.is_connected = False
        self.expiration_time = datetime.now(timezone.utc).timestamp()

    def get_bearer_token(self) -> None:
        """
        Generates new bearer token for the credentials

        Returns
        None
        """
        response = bearer_token_request(
            client_id=self.client_id,
            tenant_id=self.tenant_id,
            client_secret=self.client_secret,
        )
        self.bearer_token = response["access_token"]
        self.expiration_time = int(response["expires_on"])
        self.is_connected = True

    def check_bearer_token_validity(self) -> bool:
        """
        Provides information whether a valid bearer token is available or not. Returns true if available
        otherwise false

        Returns
        bool
        """
        if (
            self.is_connected
            and datetime.now(timezone.utc).astimezone().timestamp()
            < self.expiration_time
        ):
            return True
        else:
            return False

    def disconnect(self) -> None:
        """
        Removes bearer token from the sharepoint API class (makes it null)

        Returns
        None
        """
        self.bearer_token = None
        self.is_connected = False

    def get_all_sites(self, limit: int = None) -> List[Dict[Text, Any]]:
        """
        Gets all sites associated with the account

        limit: limits the number of site information to be returned

        Returns
        response: metadata information corresponding to all sites
        """
        url = "https://graph.microsoft.com/v1.0/sites?search=*"
        response = get_an_entity(url=url, bearer_token=self.bearer_token)
        if limit:
            response = response[:limit]
        return response

    def update_sites(
        self, site_dict: List[Dict[Text, Text]], values_to_update: Dict[Text, Any]
    ) -> None:
        """
        Updates the given sites (site_dict) with the provided values (values_to_update)
        Calls the function update_a_site for every site
        site_dict: A dictionary containing site ids of the sites which are to be updated
        values_to_update: a dictionary which will be used to update the fields of the sites

        Returns
        None
        """
        for site_entry in site_dict:
            self.update_a_site(
                site_id=site_entry["siteId"],
                values_to_update=values_to_update,
            )

    def update_a_site(self, site_id: str, values_to_update: Dict[Text, Any]) -> None:
        """
        Updates a site with given values
        site_id: GUID of the site
        values_to_update: a dictionary values which will be used to update the properties of the site

        Returns
        None
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/"
        update_an_entity(
            url=url, values_to_update=values_to_update, bearer_token=self.bearer_token
        )

    def get_lists_by_site(
        self, site_id: str, limit: int = None
    ) -> List[Dict[Text, Any]]:
        """
        Gets lists' information corresponding to a site

        site_id:  GUID of a site
        limit: limits the number of lists for which information is returned

        Returns
        response: metadata information/ fields corresponding to lists of the site
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/lists"
        response = get_an_entity(url=url, bearer_token=self.bearer_token)
        if limit:
            response = response[:limit]
        return response

    def get_all_lists(self, limit: int = None) -> List[Dict[Text, Any]]:
        """
        Gets all the lists' information assocaited with the account

        limit: puts a limit to the number of lists returned

        Returns
        response: returns metadata information regarding all the lists that have been made using that account
        """
        sites = self.get_all_sites()
        lists = []
        for site in sites:
            for list_dict in self.get_lists_by_site(site_id=site["id"].split(",")[1]):
                list_dict["siteName"] = site["name"]
                list_dict["siteId"] = site["id"].split(",")[1]
                lists.append(list_dict)
        if limit:
            lists = lists[:limit]
        return lists

    def delete_lists(self, list_dict: List[Dict[Text, Any]]) -> None:
        """
        Deletes lists for the given site ID and list ID

        list_dict: a dictionary values containing the list IDs which are to be deleted and
        their corresponding site IDs

        Returns
        None
        """
        for list_entry in list_dict:
            self.delete_a_list(site_id=list_entry["siteId"], list_id=list_entry["id"])

    def delete_a_list(self, site_id: str, list_id: str) -> None:
        """
        Deletes a list, given its list ID and its site ID

        site_id: GUID of the site in which the list is present
        list_id: GUID of the list which is to be deleted

        Returns
        None
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/lists/{list_id}"
        delete_an_entity(url=url, bearer_token=self.bearer_token)

    def update_lists(
        self, list_dict: List[Dict[Text, Text]], values_to_update: Dict[Text, Any]
    ) -> None:
        """
        Updates the given lists (list_dict) with the provided values (values_to_update)
        Calls the function update_a_list for every list
        list_dict: A dictionary containing ids of the list which are to be updated and also their site IDs
        values_to_update: a dictionary which will be used to update the fields of the lists

        Returns
        None
        """
        for list_entry in list_dict:
            self.update_a_list(
                site_id=list_entry["siteId"],
                list_id=list_entry["id"],
                values_to_update=values_to_update,
            )

    def update_a_list(
        self, site_id: str, list_id: str, values_to_update: Dict[Text, Any]
    ) -> None:
        """
        Updates a list with given values
        list_id: GUID of the list
        site_id: GUID of the site in which the list is present
        values_to_update: a dictionary values which will be used to update the properties of the list

        Returns
        None
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/lists/{list_id}/"
        update_an_entity(
            url=url, bearer_token=self.bearer_token, values_to_update=values_to_update
        )

    def create_lists(self, data: List[Dict[Text, Any]]) -> None:
        """
        Creates lists with the information provided in the data parameter
        calls create_a_list for each entry of list metadata dictionary

        data: parameter which contains information such as the site IDs where the lists would be created
        and their metadata information which will be used to create them

        Returns
        None
        """
        for entry in data:
            self.create_a_list(
                site_id=entry["siteId"],
                column=entry.get("column"),
                display_name=entry["displayName"],
                list_template=entry["list"],
            )

    def create_a_list(
        self, site_id: str, list_template: str, display_name: str, column: str = None
    ) -> None:
        """
        Creates a list with metadata information provided in the params

        site_id: GUID of the site where the list is to be created
        list_template: a string which contains the list template information (type of list)
            eg.- "{'template': 'documentLibrary'}"
        display_name: the display name of the given list, which will be displayed in the site
        column: specifies the list of columns that should be created for the list
            eg.- "[{'name': 'Author', 'text': { }},{'name': 'PageCount', 'number': { }}]"

        Returns
        None
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/lists/"
        payload = {}
        if column:
            column = eval(column)
            payload["column"] = column
        payload["displayName"] = display_name
        payload["list"] = eval(list_template)
        create_an_entity(url=url, payload=payload, bearer_token=self.bearer_token)

    def get_site_columns_by_site(
        self, site_id: str, limit: int = None
    ) -> List[Dict[Text, Any]]:
        """
        Gets columns' information corresponding to a site

        site_id:  GUID of a site
        limit: limits the number of columns for which information is returned

        Returns
        response: metadata information/ fields corresponding to columns of the site
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/columns/"
        response = get_an_entity(url=url, bearer_token=self.bearer_token)
        if limit:
            response = response[:limit]
        return response

    def get_all_site_columns(self, limit: int = None) -> List[Dict[Text, Any]]:
        """
        Gets all the columns' information associated with the account

        limit: puts a limit to the number of columns returned

        Returns
        response: returns metadata information regarding all the columns that have been made using that account
        """
        sites = self.get_all_sites()
        site_columns = []
        for site in sites:
            for site_column_dict in self.get_site_columns_by_site(
                site_id=site["id"].split(",")[1]
            ):
                site_column_dict["siteName"] = site["name"]
                site_column_dict["siteId"] = site["id"].split(",")[1]
                site_columns.append(site_column_dict)
        if limit:
            site_columns = site_columns[:limit]
        return site_columns

    def update_site_columns(
        self,
        site_column_dict: List[Dict[Text, Text]],
        values_to_update: Dict[Text, Any],
    ) -> None:
        """
        Updates the given columns (site_column_dict) with the provided values (values_to_update)
        Calls the function update_a_site_column for every column

        site_column_dict: A dictionary containing ids of the column which are to be updated and
        also their site IDs
        values_to_update: a dictionary which will be used to update the fields of the columns

        Returns
        None
        """
        for site_column_entry in site_column_dict:
            self.update_a_site_column(
                site_id=site_column_entry["siteId"],
                column_id=site_column_entry["id"],
                values_to_update=values_to_update,
            )

    def update_a_site_column(
        self, site_id: str, column_id: str, values_to_update: Dict[Text, Any]
    ):
        """
        Updates a column with given values

        column_id: GUID of the column
        site_id: GUID of the site in which the column is present
        values_to_update: a dictionary values which will be used to update the properties of the column

        Returns
        None
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/columns/{column_id}"
        update_an_entity(
            url=url, values_to_update=values_to_update, bearer_token=self.bearer_token
        )

    def delete_site_columns(self, column_dict: List[Dict[Text, Any]]) -> None:
        """
        Deletes columns for the given site ID and column ID

        column_dict: a dictionary values containing the column IDs which are to be deleted and
        their corresponding site IDs

        Returns
        None
        """
        for column_entry in column_dict:
            self.delete_a_site_columns(
                site_id=column_entry["siteId"], column_id=column_entry["id"]
            )

    def delete_a_site_columns(self, site_id: str, column_id: str) -> None:
        """
        Deletes a column, given its column ID and its site ID

        site_id: GUID of the site in which the column is present
        column_id: GUID of the column which is to be deleted

        Returns
        None
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/columns/{column_id}"
        delete_an_entity(url=url, bearer_token=self.bearer_token)

    def create_site_columns(self, data: List[Dict[Text, Any]]) -> None:
        """
        Creates columns with the information provided in the data parameter
        calls create_a_site_column for each entry of column metadata dictionary

        data: parameter which contains information such as the site IDs where the columns would be created
        and their metadata information which will be used to create them

        Returns
        None
        """
        for entry in data:
            self.create_a_site_column(
                site_id=entry["siteId"],
                enforce_unique_values=entry.get("enforceUniqueValues"),
                hidden=entry.get("hidden"),
                indexed=entry.get("indexed"),
                name=entry["name"],
                text=entry.get("text"),
            )

    def create_a_site_column(
        self,
        site_id: str,
        enforce_unique_values: bool,
        hidden: bool,
        indexed: bool,
        name: str,
        text: str = None,
    ) -> None:
        """
        Creates a list with metadata information provided in the params

        site_id: GUID of the site where the column is to be created
        enforced_unique_values: if true, no two list items may have the same value for this column
        hidden: specifies whether the column is displayed in the user interface
        name: the API-facing name of the column as it appears in the fields on a listItem.
        text: details regarding the text values in the column

        Returns
        None
        """

        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/columns/"
        payload = {}
        if text:
            text = eval(text)
            payload["text"] = text
        payload["name"] = name
        if enforce_unique_values is not None:
            payload["enforceUniqueValues"] = enforce_unique_values
        if hidden is not None:
            payload["hidden"] = hidden
        if indexed is not None:
            payload["indexed"] = indexed
        create_an_entity(url=url, payload=payload, bearer_token=self.bearer_token)

    def get_items_by_sites_and_lists(
        self, site_id: str, list_id: str, limit: int = None
    ) -> List[Dict[Text, Any]]:
        """
        Gets items' information corresponding to a site and a list

        site_id:  GUID of a site
        list_id: GUID of a list
        limit: limits the number of columns for which information is returned

        Returns
        response: metadata information/ fields corresponding to list-items of the site
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/lists/{list_id}/items?expand=fields&select=*"
        response = get_an_entity(url=url, bearer_token=self.bearer_token)
        if limit:
            response = response[:limit]
        return response

    def get_all_items(self, limit: int = None) -> List[Dict[Text, Any]]:
        """
        Gets all the items' information associated with the account

        limit: puts a limit to the number of items returned

        Returns
        response: returns metadata information regarding all the items that are associated with that account
        """
        sites = self.get_all_sites()
        items = []
        for site in sites:
            site_id = site["id"].split(",")[1]
            for sharepoint_list in self.get_lists_by_site(site_id=site_id):
                for item_dict in self.get_items_by_sites_and_lists(
                    site_id=site["id"].split(",")[1], list_id=sharepoint_list["id"]
                ):
                    item_dict["siteName"] = site["name"]
                    item_dict["siteId"] = site["id"].split(",")[1]
                    item_dict["listId"] = sharepoint_list["id"]
                    item_dict["list_name"] = sharepoint_list["displayName"]
                    items.append(item_dict)
        if limit:
            items = items[:limit]
        return items

    def update_items(
        self, item_dict: List[Dict[Text, Text]], values_to_update: Dict[Text, Any]
    ) -> None:
        """
        Updates the given items (item_dict) with the provided values (values_to_update)
        Calls the function update_a_item for every column

        item_dict: A dictionary containing ids of the list-items which are to be updated and
        also their site IDs
        values_to_update: a dictionary which will be used to update the fields of the items

        Returns
        None
        """
        for item_entry in item_dict:
            self.update_an_item(
                site_id=item_entry["siteId"],
                list_id=item_entry["listId"],
                item_id=item_entry["id"],
                values_to_update=values_to_update,
            )

    def update_an_item(
        self,
        site_id: str,
        list_id: str,
        item_id: str,
        values_to_update: Dict[Text, Any],
    ):
        """
        Updates an item with given values

        item_id: GUID of the column
        list_id: GUID of the list
        site_id: GUID of the site in which the list is present
        values_to_update: a dictionary values which will be used to update the properties of the list-item

        Returns
        None
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/lists/{list_id}/items/{item_id}"
        update_an_entity(
            url=url, values_to_update=values_to_update, bearer_token=self.bearer_token
        )

    def delete_items(self, item_dict: List[Dict[Text, Any]]) -> None:
        """
        Deletes items for the given site ID and list ID

        item_dict: a dictionary values containing the item IDs which are to be deleted,
        their corresponding site IDs and their list IDs

        Returns
        None
        """
        for item_entry in item_dict:
            self.delete_an_item(
                site_id=item_entry["siteId"],
                list_id=item_entry["listId"],
                item_id=item_entry["id"],
            )

    def delete_an_item(self, site_id: str, list_id: str, item_id: str) -> None:
        """
        Deletes an item, given its item ID, its site ID and its list ID

        list_id: GUID of the list in which the site is present
        site_id: GUID of the site in which the list is present
        item_id: GUID of the item which is to be deleted

        Returns
        None
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/lists/{list_id}/items/{item_id}"
        delete_an_entity(url=url, bearer_token=self.bearer_token)

    def create_items(self, data: List[Dict[Text, Any]]) -> None:
        """
        Creates items with the information provided in the data parameter
        calls create_an_item for each entry of item metadata dictionary

        data: parameter which contains information such as the site IDs and list IDs where the items
        would be created and their metadata information which will be used to create them

        Returns
        None
        """
        for entry in data:
            self.create_an_item(
                site_id=entry["siteId"],
                list_id=entry["listId"],
                fields=entry.get("fields"),
            )

    def create_an_item(self, site_id: str, list_id: str, fields: str) -> None:
        """
        Creates an item with metadata information provided in the params

        site_id: GUID of the site where the list id present
        list_id: GUID of the list where the item is to be created
        fields: The values of the columns set on this list item.

        Returns
        None
        """
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/lists/{list_id}/items/"
        payload = {}
        if fields:
            payload["fields"] = eval(fields)
        create_an_entity(url=url, payload=payload, bearer_token=self.bearer_token)



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __description__ as description
from .__about__ import __version__ as version

try:
    from .langchain_embedding_handler import LangchainEmbeddingHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "LangChain Embedding"
name = "langchain_embedding"
type = HANDLER_TYPE.ML
permanent = False

__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error"]



import time
import requests
from typing import Dict, List
from requests_toolbelt.multipart.encoder import MultipartEncoder

from mindsdb.utilities import log


logger = log.getLogger(__name__)

# TODO: move to config
BASE_URL = "https://api.twelvelabs.io/v1.1"
DEFAULT_ENGINE = "marengo2.5"
DEFAULT_WAIT_DURATION = 5


class TwelveLabsAPIClient:
    def __init__(self, api_key):
        self.api_key = api_key
        self.headers = {
            'Content-Type': 'application/json',
            'x-api-key': self.api_key
        }

    def create_index(self, index_name: str, index_options: List[str], engine_id: str = None, addons: List[str] = None) -> str:
        """
        Create an index.

        """
        body = {
            "index_name": index_name,
            "engine_id": engine_id if engine_id else DEFAULT_ENGINE,
            "index_options": index_options,
            "addons": addons,
        }

        result = self._submit_request(
            method="POST",
            endpoint="indexes",
            data=body,
        )

        logger.info(f"Index {index_name} successfully created.")
        return result['_id']

    def get_index_by_name(self, index_name: str) -> str:
        """
        Get an index by name.

        """
        params = {
            "index_name": index_name,
        }

        result = self._submit_request(
            method="GET",
            endpoint="indexes",
            data=params,
        )

        data = result['data']
        return data[0]['_id'] if data else None

    def create_video_indexing_tasks(self, index_id: str, video_urls: List[str] = None, video_files: List[str] = None) -> List[str]:
        """
        Create video indexing tasks.

        """
        task_ids = []

        if video_urls:
            logger.info("video_urls has been set, therefore, it will be given precedence.")
            logger.info("Creating video indexing tasks for video urls.")

            for video_url in video_urls:
                task_ids.append(
                    self._create_video_indexing_task(
                        index_id=index_id,
                        video_url=video_url
                    )
                )

        elif video_files:
            logger.info("video_urls has not been set, therefore, video_files will be used.")
            logger.info("Creating video indexing tasks for video files.")

            for video_file in video_files:
                task_ids.append(
                    self._create_video_indexing_task(
                        index_id=index_id,
                        video_file=video_file
                    )
                )

        return task_ids

    def _create_video_indexing_task(self, index_id: str, video_url: str = None, video_file: str = None) -> str:
        """
        Create a video indexing task.

        """
        body = {
            "index_id": index_id,
        }

        if video_url:
            body['video_url'] = video_url
        elif video_file:
            body['video_file'] = video_file

        result = self._submit_multi_part_request(
            method="POST",
            endpoint="tasks",
            data=body,
        )

        task_id = result['_id']
        logger.info(f"Created video indexing task {task_id} for {video_url if video_url else video_file} successfully.")
        return task_id

    def poll_for_video_indexing_tasks(self, task_ids: List[str]) -> None:
        """
        Poll for video indexing tasks to complete.

        """
        for task_id in task_ids:
            logger.info(f"Polling status of video indexing task {task_id}.")
            is_task_running = True

            while is_task_running:
                task = self._get_video_indexing_task(task_id=task_id)
                status = task['status']
                logger.info(f"Task {task_id} is in the {status} state.")

                wait_durtion = task['process']['remain_seconds'] if 'process' in task else DEFAULT_WAIT_DURATION

                if status in ('pending', 'indexing', 'validating'):
                    logger.info(f"Task {task_id} will be polled again in {wait_durtion} seconds.")
                    time.sleep(wait_durtion)

                elif status == 'ready':
                    logger.info(f"Task {task_id} completed successfully.")
                    is_task_running = False

                else:
                    logger.error(f"Task {task_id} failed with status {task['status']}.")
                    # TODO: update Exception to be more specific
                    raise Exception(f"Task {task_id} failed with status {task['status']}.")

        logger.info("All videos indexed successfully.")

    def _get_video_indexing_task(self, task_id: str) -> Dict:
        """
        Get a video indexing task.

        """
        result = self._submit_request(
            method="GET",
            endpoint=f"tasks/{task_id}",
        )

        logger.info(f"Retrieved video indexing task {task_id} successfully.")
        return result

    def search_index(self, index_id: str, query: str, search_options: List[str]) -> Dict:
        """
        Search an index.

        """
        body = {
            "index_id": index_id,
            "query": query,
            "search_options": search_options
        }

        data = []
        result = self._submit_request(
            method="POST",
            endpoint="search",
            data=body,
        )
        data.extend(result['data'])

        while('next_page_token' in result['page_info']):
            result = self._submit_request(
                method="GET",
                endpoint=f"search/{result['page_info']['next_page_token']}"
            )
            data.extend(result['data'])

        logger.info(f"Search for index {index_id} completed successfully.")
        return data

    def _submit_request(self, endpoint: str, headers: Dict = None, data: Dict = None, method: str = "GET") -> Dict:
        """
        Submit a request to the Twelve Labs API.

        """
        url = f"{BASE_URL}/{endpoint}"

        headers = headers if headers else self.headers

        if method == "GET":
            response = requests.get(
                url=url,
                headers=headers,
                params=data if data else {},
            )

        elif method == "POST":
            response = requests.post(
                url=url,
                headers=headers,
                json=data if data else {},
            )

        else:
            raise Exception(f"Method {method} not supported yet.")

        result = response.json()
        if response.status_code in (200, 201):
            logger.info("API request was successful.")
            return result
        else:
            logger.error(f"API request has failed: {result['message']}")
            # TODO: update Exception to be more specific
            raise Exception(f"API request has failed: {result['message']}")

    def _submit_multi_part_request(self, endpoint: str, headers: Dict = None, data: Dict = None, method: str = "POST") -> Dict:
        """
        Submit a multi-part request to the Twelve Labs API.

        """
        url = f"{BASE_URL}/{endpoint}"

        headers = headers = headers if headers else self.headers

        multipart_data = MultipartEncoder(fields=data)
        headers['Content-Type'] = multipart_data.content_type

        if method == "POST":
            response = requests.post(
                url=url,
                headers=headers,
                data=multipart_data if multipart_data else {},
            )

        else:
            raise Exception(f"Method {method} not supported yet.")

        result = response.json()
        if response.status_code in (200, 201):
            logger.info("API request was successful.")
            return result
        else:
            logger.error(f"API request has failed: {result['message']}")
            # TODO: update Exception to be more specific
            raise Exception(f"API request has failed: {result['message']}")
